# Palm::ZText.pm
#
# Perl class for dealing with Palm ZText databases used by gutenpalm.
# 
# 	Copyright (C) 2001 Geoffrey Reedy.
# 	You may distribute this file in under the terms of the
#	GNU General Public License, as found in the COPYING file.
#
# $Id: ZText.pm,v 1.3 2001/06/15 04:25:23 vader21 Exp $ 

use strict;
package Palm::ZText;
use Palm::Raw();
use Data::Dumper;
use Compress::Zlib;
use Carp;
use vars qw( $VERSION @ISA );

$VERSION = (qw( $Revision: 1.3 $ )[1]);
@ISA = qw( Palm::Raw );

# constants
my $ZTXT_CREATOR = "GPlm";
my $ZTXT_TYPE = "zTXT";
my $ZTXT_VERSION = 269;
my $ZTXT_RECORDSIZE = 8192;
my $ZTXT_MAX_BMRK_LENGTH = 20;


my $recordnum = 0;

my $headerpackstring = 
	    "n" . # version (UInt16)
	    "n" . # numRecords (UInt16
	    "N" . # size (UInt32)
	    "n" . # recordSize (UInt16)
	    "n" . # numBookmarks (Uint16)
	    "n" . # bookmarkRecord (Uint16)
	    "n" . # numAnnotations (UInt16)
	    "n" . # annotationRecord (UInt16)
	    "C" . # randomAccess (UInt8)
	    "n" . # storedPrefs (UInt16)
	    "n" . # buildNumber (UInt16)
	    "x9"; # pad to 32 bytes

=head1 NAME

Palm::ZText - Handler for Palm ZText databases used by gutenpalm

=head1 SYNOPSIS

    use Palm::ZText;

=head1 DESCRIPTION

The ZText PDB handler is a helper class for the Palm::PDB module.  It parses
ZText databases used by the gutenpalm document reader.

=cut

sub import 
{
    &Palm::PDB::RegisterPDBHandlers(__PACKAGE__,
	[ "GPlm", "zTXT" ]
	);
}

=head2 new 

    $pdb = new Palm::ZText;

Create a new ZText PDB object.  Use this to create a ZText PDB from scratch.
    
    $pdb = new Palm::ZText "filename.txt";

Load a ZText PDB from a file.

=cut

sub new
{
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    my $record0;
    my $filename;

# required initialization

    if ($filename = shift) {
	$self->Load($filename);
	return $self;
    }

    $self->{name} = "";
    $self->{creator} = "GPlm";
    $self->{type} = "zTXT";
    $self->{attributes}{resource} = 0; # explicitly say we are a DB


# Things that matter for a ZText document are here

    $self->{records} = []; # use empty list of records;

# Set up the ZText header with appropriate information
    $record0 = $self->append_Record;
    %$record0 = ( %$record0,
	type => 'header',
	version => $ZTXT_VERSION,		# zTXT format revision
	numRecords => 0,			# number of data records
	size => 0,				# size of compressed data
	recordSize => $ZTXT_RECORDSIZE,		# size of each record
	numBookmarks => 0,			# number of bookmarks
	bookmarkRecord => 0,			# ID of bookmark record
	numAnnotations => 0,			# number of annotations
	annotationRecord => 0,			# ID of annotation record
	randomAccess => 0,			# if comp allows random access
	storedPrefs => 0,			# per document prefs
	buildNumber => 0,			# build counter
	majorsort => 0,
	minorsort => 0,
	id => 0
    );

    # Initialize the unused fields so that Palm::PDB isn't noisy when
    # using perl -w
    $self->{appinfo} = "";
    $self->{sort} = "";
    $self->{modnum} = 0;
    $self->{uniqueIDseed} = 0;
    $self->{"2NULs"} = "";

    $recordnum = 0;  # used by ParseRecord to figure out which record is which

    return $self;
}

=head2 Load

    $pdb->Load("filename");

Loads the PDB file given as a parameter into the ZText object.

=cut

sub Load
{
    my $self = shift;
    if (defined $self->{istream}) {
	delete $self->{istream};
    }
    $self->{istream} = inflateInit();
    $recordnum = 0;
    return $self->SUPER::Load(@_);
}

# Since the ZText format uses multiple records to contain a single logical
# block of information, things get a little weird from this point on I think

sub ParseRecord
{
    my $self = shift;
    my %record = @_;
    my $totalrecords;

    $record{recordnum} = $recordnum;

    if ( $recordnum == 0 ) {
# If this is record 0, then it contains the header for the document

	$record{type} = 'header';
	$record{majorsort} = 1;
	$record{minorsort} = 1;

	@record{'version','numRecords','size','recordSize',
			'numBookmarks','bookmarkRecord','numAnnotations',
			'annotationRecord','randomAccess','storedPrefs',
			'buildNumber'
			} = unpack($headerpackstring, $record{data});
	delete $record{data};
    } elsif ( $recordnum <= $self->{records}[0]{numRecords} ) {
# this is a record containing text


	$record{type} = 'text';
	$record{majorsort} = 2;
	$record{minorsort} = $recordnum;
	$record{text} = $self->{istream}->inflate($record{data});
	$record{textlength} = length $record{text};

	delete $record{data};
    } elsif ( $recordnum == $self->{records}[0]{bookmarkRecord} ) {
# this is the bookmark record

	$record{type} = 'bookmarks';
	$record{majorsort} = 3;
	$record{minorsort} = 1;
	$record{bookmarks} = [];
	for (1..$self->{records}[0]{numBookmarks}) {
	    my ($bmname, $bmoffset);
	    ($bmoffset, $bmname) = 
		unpack ("NZ$ZTXT_MAX_BMRK_LENGTH", $record{data});
	    $record{data} = substr ($record{data}, $ZTXT_MAX_BMRK_LENGTH+4);
	    push (@{$record{bookmarks}} ,
		{'name' => $bmname, 'offset' => $bmoffset});
	}
	delete $record{data};

    } elsif ( $recordnum == $self->{records}[0]{annotationRecord} ) {
# this is the annotation record

	my $annonum = 1;

	$record{type} = 'annotations';
	$record{majorsort} = 4;
	$record{minorsort} = 1;
	$record{annotations} = [];
	for (1..$self->{records}[0]{numAnnotations}) {
	    my ($anname, $anoffset, $anrecordnum, $anrecord);
	    ($anoffset, $anname) = 
		unpack("NZ$ZTXT_MAX_BMRK_LENGTH", $record{data});
	    $record{data} = substr ($record{data},$ZTXT_MAX_BMRK_LENGTH+4);
	    $anrecordnum = $self->{records}[0]{annotationRecord} + $annonum;
	    $anrecord = \$self->{records}[$anrecordnum];
	    push (@{$record{annotations}}, 
		{'name' => $anname, 'offset' => $anoffset, 
		 'recordnum' => $anrecordnum, 'record' => $anrecord});
	    $annonum++;
	}
	delete $record{data};
    } else { # must be an annotation
	my ($annotationnum, $indexentry, $annotationindex);
	$record{type} = 'annotation';
	$annotationindex = $self->{records}[0]{annotationRecord};
	$annotationnum = $recordnum - $annotationindex - 1;
	$indexentry = 
	    \$self->{records}[$annotationindex]{annotations}[$annotationnum];
	$record{majorsort} = 5;
	$record{minorsort} = $annotationnum;
	$record{annotationnum} = $annotationnum;
	$record{indexentry} = $indexentry;
	# get rid of the trailing null
	chop $record{data};
    }

    $recordnum++;
    return \%record;
}

sub dump
{
    my $self = shift;

    return Dumper($self);
}

=head2 Write

    $pdb->Write("filename")

Writes the ZText document to the given filename.

=cut

sub Write
{

    my $self = shift;

    if (defined $self->{dstream}) {
	delete $self->{dstream};
    }
    $self->{dstream} = deflateInit;

# FIXME for now assume that everything comes out in the right order
# we really need to reorder the records like this
# header, text blocks, bookmarks, annotations
    @{$self->{records}} = sort { $a->{majorsort} <=> $b->{majorsort} 
	   or $a->{minorsort} <=> $b->{minorsort} } @{$self->{records}};

    return $self->SUPER::Write(@_);
}

sub PackRecord
{

    my $self = shift;
    my $record = shift;
    my $data;

    #print Dumper($record);
    if ($record->{type} eq 'header') {
	return pack($headerpackstring, 
	    @{%$record}{'version','numRecords','size',
			'recordSize','numBookmarks',
			'bookmarkRecord','numAnnotations',
			'annotationRecord','randomAccess',
			'storedPrefs','buildNumber'});
    } elsif ($record->{type} eq 'text') {
	$data = $self->{dstream}->deflate($record->{text});
	$data .= $self->{dstream}->flush(Z_FULL_FLUSH);
	return $data;
    } elsif ($record->{type} eq 'bookmarks') {
	$data = '';
	my $bookmark;
	foreach $bookmark ($record->{bookmarks}) {
	   $data .= pack ("NZ$ZTXT_MAX_BMRK_LENGTH", 
		    ($bookmark->{offset}, $bookmark->{name}));
	}
	return $data;
    } elsif ($record->{type} eq 'annotations') {
	$data = '';
	my $annotation;
	foreach $annotation ($record->{annotations}) {
	   $data .= pack ("NZ$ZTXT_MAX_BMRK_LENGTH", 
		    ($annotation->{offset}, $annotation->{name}));
	}
	return $data;
    } elsif ($record->{type} eq 'annotation') {
	return $record->{data}."\0";
    }
}

=head2 textfile

    $pdb->textfile("filename");

Set the text contents of the ZText document to the contents of the given
file.

=cut

sub textfile
{
    my $self = shift;
    my $filename = $_[0];
    my $text;
    local $/;
    undef $/;

    open(FILE,"<",$filename) or croak "Couldn't open $filename"; # safer

    $text = <FILE>;

    $self->text($text);

    close FILE;
}

=head2 text

    $text = $pdb->text;

Return the text contents of the ZText database.

    $pdb->text $text;

Set the text contents of the ZText database.  This usage also clears all
of the bookmarks and annotations, as they really wouldn't be usefull with
the new text.

    $pdb->text @lines;

Set the text contents of the ZText database.  This method accepts a list
of lines that will be the ZText contents and otherwise behaves axactly as
the above usage.

=cut

sub text
{
    my $self = shift;
    if (@_) {
	my $text = join("\n", @_);

	for (0..$#{$self->{records}}) {
	    delete $self->{records}[$_] 
		unless $self->{records}[$_]->{type} eq 'header';
	}

	# Should always be the first record but just to be safe...
	my $header = (grep {$_->{type} eq 'header'} @{$self->{records}} )[0];

	$header->{numRecords} = 0;
	$header->{recordSize} = $ZTXT_RECORDSIZE;
	$header->{size} = length $text;
	$header->{numBookmarks} = 0;
	$header->{bookmarkRecord} = 0;
	$header->{numAnnotations} = 0;
	$header->{annotationRecord} = 0;
	$header->{randomAccess} = 1;
	$header->{storedPrefs} = 0;
	$header->{buildNumber} = 0;
	
	while (length $text) {
	    my $segment = substr($text,0,$ZTXT_RECORDSIZE,'');
	    my $record = $self->append_Record;

	    $record->{type} = 'text';
	    $record->{majorsort} = 2;
	    $record->{minorsort} = $header->{numRecords}++;
	    $record->{id} = $header->{numRecords};
	    $record->{text} = $segment;
	}
    } else { 
	my $text;
	my $record;
	foreach $record ($self->{records}) {
	    $text .= $record->{text} if $record->{type} eq 'text';
	}
	return $text;
    }
}

=head2 bookmarks

    @bookmarks = $pdb->bookmarks;

Returns list of all bookmarks currently defined in the database.  If there
are no bookmarks, returns an empty list.  Each element of the returned list
is a hashref with keys name and offset.  Name contains the name of the
bookmark that is displayed in the bookmark menu, and offset contains the
offset into the text that the bookmarks is at.

    $pdb->bookmarks @bookmarks;

(Re)sets the list of bookmarks to those in the given list.  Each member of
the list should be a hashref in the format returned with the above usage.
Note: When writing the database, constraints of the format may cause the name
of each bookmark to be truncated.  Also, this method replaces the list of
bookmarks, to append bookmarks use the appendbookmarks method described below.

=cut

sub bookmarks
{
    my $self = shift;
    my $header = (grep {$_->{type} eq 'header'} @{$self->{records}})[0];

    if (@_) {
	my $record;
	if ($header->{bookmarkRecord} == 0) {
	    $record = $self->append_Record;
	    $record->{type} = 'bookmarks';
	    $record->{majorsort} = 3;
	    $record->{minorsort} = 1;
	    $header->{bookmarkRecord} = $#{$self->{records}};
	} else {
	    $record = $self->{records}[$header->{bookmarkRecord}];
	}
	# XXX eventually check to see that the bookmarks passed are valid
	$header->{numBookmarks} = $#_;
	$record->{bookmarks} = \@_;
    } else {
	if ($header->{bookmarkRecord} == 0) {
	    return qw();
	} else {
	    return $self->{records}[$header->{bookmarkRecord}]{bookmarks};
	}
    }
}

=head2 appendbookmarks

    $pdb->appendbookmarks($bookmark1, $bookmark2, ...)
    $pdb->appendbookmarks(@bookmarks)

Appends to the list of bookmarks in the database.  Each argument should
be a hashref with keys name and offset with name being the name of the 
bookmark, and offset being the offset in the text where the bookmarks is
located.

=cut

sub appendbookmarks
{
    my $self = shift;
    my $header = (grep {$_->{type} eq 'header'} @{$self->{records}})[0];
    my $record;

    if ($header->{bookmarkRecord} == 0) { 
	$record = $self->append_Record;
	$record->{type} = 'bookmarks';
	$record->{majorsort} = 3;
	$record->{minorsort} = 1;
	$record->{bookmakrs} = [];
	$header->{bookmarkRecord} = $#{$self->{records}};
    } else {
	$record = $self->{records}[$header->{bookmarkRecord}];
    }

    push (@{$record->{bookmarks}}, @_);

}

=head2 name
    
    $name = $pdb->name;

Returns name of document as a string.

    $pdb->name("name");

Sets the name of the document to "name".

=cut

sub name
{
    my $self = shift;
    if (@_) {
	$self->{name} = $_[0];
    } else {
	return $self->{name};
    }
}

1;
__END__

=head1 AUTHOR

Geoff Reedy E<lt>vader21@imsa.eduE<gt>

=head1 SEE ALSO

L<Palm::PDB(3)>

=cut
